% ACCV 2014 - Minimal solution for computing pairs of lines in non-central cameras:
% J. Bermudez-Cameo, J.P. Barreto, G. Lopez-Nicolas, J.J. Guerrero
% Supplementary Material - ID394
% ==============================================================================
% Copyright (C) 2014 Jesus Bermudez-Cameo.
% This code has been implemented by Jesus Bermudez-Cameo and supervised 
% by J.P. Barreto, G.Lopez-Nicolas and J.J. Guerrero

close all;
clear;
dbstop if error;
iptsetpref('ImshowBorder','tight');

if isunix
    path(path,'sphereProjection/');
    path(path,'P5LinearTeller/');
    path(path,'P5Junction/');
    path(path,'P5Parallels/');
    path(path,'plotTools/');
    path(path,'P5operators/');    
else
    path(path,'sphereProjection\');
    path(path,'P5LinearTeller\');
    path(path,'P5Junction\');
    path(path,'P5Parallels\');
    path(path,'plotTools\');
    path(path,'P5operators\');    
end


%% Simulation Configuration
configFlag = 'A';       % Lines configuration
lengthLinesFlag = 0;    % 0 for long lines ; 1 for short lines
minimalSetFlag = 1;     % 1 minimal set ; 0 all points

switch(configFlag)
    case 'A',
        proposalFlag = 0; % 0 Junction ; 1 Parallel
        if lengthLinesFlag==1
            load configAshortLines.mat;
        else
            load configAlongLines.mat;
        end        
    case 'B',
        proposalFlag = 0; % 0 Junction ; 1 Parallel
        if lengthLinesFlag==1
            load configBshortLines.mat;
        else
            load configBlongLines.mat;
        end
    case 'C',
        proposalFlag = 1;  % 0 Junction ; 1 Parallel        
        if lengthLinesFlag==1
            load configCshortLines.mat;
        else
            load configClongLines.mat;
        end        
end


L_GT = twoPointsSpan4([XLini;ones(1,size(XLini,2))],[XLend;ones(1,size(XLend,2))]);
M_GT = twoPointsSpan4([XMini;ones(1,size(XMini,2))],[XMend;ones(1,size(XMend,2))]);

nAttempts = size(L_GT,2);
%% Calibration and configuration of the non-central system
d = 1.8;
rSph = 1.2;

mRows = 768;
nColumns = 1024;

u0 = nColumns/2;
v0 = mRows/2;

fx = 500;
fy = 500;

K = [[fx 0 u0];
     [0 fy v0];
     [0  0  1]];

%% Camara
tCam = [0;0;d];
RCam = [[1,0,0];[0,-1,0];[0,0,-1]];
TCam = [[RCam tCam];[zeros(1,3) 1]];


figure(1);
hold on;
axis equal;
plotSphericalSystem(rSph,tCam,RCam);


tParamLine = 0:0.01:1;
DeltaXL = XLend-XLini;
DeltaXM = XMend-XMini;

nPoints = numel(tParamLine);

qLGT = cell(1,nAttempts);qMGT = cell(1,nAttempts);
xIArrayL = cell(1,nAttempts);xIArrayM = cell(1,nAttempts);
for k = 1:nAttempts
        
    XL = [bsxfun(@plus,XLini(:,k),DeltaXL(:,k)*tParamLine);ones(1,nPoints)];
    XM = [bsxfun(@plus,XMini(:,k),DeltaXM(:,k)*tParamLine);ones(1,nPoints)];

    plot3(XL(1,:),XL(2,:),XL(3,:),'b','LineWidth',2);
    plot3(XM(1,:),XM(2,:),XM(3,:),'r','LineWidth',2);
    
    
    [qLGT{k}, xIArrayL{k}] = forwardSphereProjection(XL,TCam,K,d,rSph);
    [qMGT{k}, xIArrayM{k}] = forwardSphereProjection(XM,TCam,K,d,rSph);                    
end

sigmaError = [0;0.1;0.3;0.5;1;1.5;2];



distvLLinAcum = zeros(nAttempts,numel(sigmaError));distdLLinAcum = zeros(nAttempts,numel(sigmaError));
distvMLinAcum = zeros(nAttempts,numel(sigmaError));distdMLinAcum = zeros(nAttempts,numel(sigmaError));

distvLProposalAcum = zeros(nAttempts,numel(sigmaError));distdLProposalAcum = zeros(nAttempts,numel(sigmaError));
distvMProposalAcum = zeros(nAttempts,numel(sigmaError));distdMProposalAcum = zeros(nAttempts,numel(sigmaError));


for kNoiseIndex = 1:numel(sigmaError)
    sigmaL = sigmaError(kNoiseIndex);
    sigmaM = sigmaError(kNoiseIndex);


    for kNoise = 1:nAttempts
                
        qL = qLGT{kNoise} + [randn(2,nPoints)*sigmaL;zeros(1,nPoints)];
        qM = qMGT{kNoise} + [randn(2,nPoints)*sigmaM;zeros(1,nPoints)];
        
        %% Backprojecting the rays
        [~, ~, ~, mMinus, ~, bMinus] = singleBackSphereProjection(qL,d,rSph,K);
        [xiL, xiBarL] = twoPointsSpan(ones(1,size(mMinus,2)),RCam*mMinus,ones(1,size(mMinus,2)),RCam*bMinus);

        XiL = [xiL;xiBarL];
        XiL = XiL(:,sum(imag(XiL)==0)==6);        
        xiL = XiL(1:3,:); xiBarL = XiL(4:6,:);
        
        %% Backprojecting the rays
        [~, ~, ~, mMinus, ~, bMinus] = singleBackSphereProjection(qM,d,rSph,K);
        [xiM, xiBarM] = twoPointsSpan(ones(1,size(mMinus,2)),RCam*mMinus,ones(1,size(mMinus,2)),RCam*bMinus);

        XiM = [xiM;xiBarM];
        XiM = XiM(:,sum(imag(XiM)==0)==6);        
        xiM = XiM(1:3,:); xiBarM = XiM(4:6,:);

        
        localnPoints = min(size(XiL,2),size(XiM,2));        
        if(minimalSetFlag==1)
            iii = [floor(localnPoints/4) floor(localnPoints/2) ceil(localnPoints*3/4)];
            iiii = [iii localnPoints];
        else
            iii = 1:localnPoints;
            iiii = 1:localnPoints;
        end
        
        %% Teller's linear approach        
        [LLin1,LLin2,FL,GL]= tellerLineExtraction(xiL(:,iiii),xiBarL(:,iiii));
        LLin = getCorrectTellerLine(LLin1,LLin2);    
        [MLin1,MLin2,FM,GM]= tellerLineExtraction(xiM(:,iiii),xiBarM(:,iiii));
        MLin = getCorrectTellerLine(MLin1,MLin2);

        [distvLLin, ~, distdLLin, LVotes] = getPluckerLineErrors(LLin,L_GT(:,kNoise));
        [distvMLin, ~, distdMLin, MVotes] = getPluckerLineErrors(MLin,M_GT(:,kNoise));
                
        iL = find(LVotes==max(LVotes));
        iM = find(MVotes==max(MVotes));        
        if numel(iL)~=1, if numel(iL)>1, iL = iL(1);end, end
        if numel(iM)~=1, if numel(iM)>1, iM = iM(1);end, end
        
        distvLLinAcum(kNoise,kNoiseIndex) = distvLLin(iL);
        distdLLinAcum(kNoise,kNoiseIndex) = distdLLin(iL);
        
        distvMLinAcum(kNoise,kNoiseIndex) = distvMLin(iM);
        distdMLinAcum(kNoise,kNoiseIndex) = distdMLin(iM);    
        
        
        if proposalFlag==1
            [LPar,MPar] = twoParallelLinesExtraction(xiL(:,iii),xiBarL(:,iii),xiM(:,iii),xiBarM(:,iii));
            LProposal = LPar; MProposal = MPar;
        else
            [Ljt,Mjt,lambda,mu,realFlag] = linesJunctionExtraction(XiL(:,iii),XiM(:,iii));
            LProposal = Ljt(:,realFlag); MProposal = Mjt(:,realFlag);            
        end
        
       
        [distvLProposal, ~, distdLProposal, LVotes] = getPluckerLineErrors(LProposal,L_GT(:,kNoise));
        [distvMProposal, ~, distdMProposal, MVotes] = getPluckerLineErrors(MProposal,M_GT(:,kNoise));

         
        iL = find(LVotes==max(LVotes));
        iM = find(MVotes==max(MVotes));        
        if numel(iL)~=1, if numel(iL)>1, iL = iL(1);end, end
        if numel(iM)~=1, if numel(iM)>1, iM = iM(1);end, end        
        
        distvLProposalAcum(kNoise,kNoiseIndex) = distvLProposal(iL);
        distdLProposalAcum(kNoise,kNoiseIndex) = distdLProposal(iL);
        
        distvMProposalAcum(kNoise,kNoiseIndex) = distvMProposal(iM);
        distdMProposalAcum(kNoise,kNoiseIndex) = distdMProposal(iM);    
        
        
                
    end
    

    
end

%% Plot the Analysis

sigmaExt = [sigmaError;sigmaError];
sigmaExtText = repmat({'       0.0' '       0.1' '       0.3' '       0.5' '       1.0' '       1.5' '       2.0'},1,2);


if proposalFlag==1
    LinProposal = [repmat({'Lin'},1,7),repmat({'Par'},1,7)];
else
    LinProposal = [repmat({'Lin'},1,7),repmat({'2-jt'},1,7)];
end

distv = [[distvLLinAcum distvLProposalAcum];
         [distvMLinAcum distvMProposalAcum]];                          

figure(2);
set(gca,'fontSize',15);
set(gca,'LooseInset',get(gca,'TightInset'));
boxplot(distv/pi*180,{sigmaExtText,LinProposal},'factorgap',[5 2],'labelverbosity','minor','datalim',[-Inf Inf]);
if minimalSetFlag==1 && lengthLinesFlag==0
    title('20m line length - Minimal Set');
end
if minimalSetFlag==1 && lengthLinesFlag==1
    title('1m line length - Minimal Set');
end

if minimalSetFlag==0 && lengthLinesFlag==0
    title('20m line length - All points');
end

if minimalSetFlag==0 && lengthLinesFlag==1
    title('1m line length - All points');
end

xlabel('gaussian error (pixels)');
ylabel('direction error (deg)');




distd = [[distdLLinAcum distdLProposalAcum];
         [distdMLinAcum distdMProposalAcum]]; 
figure(3);
set(gca,'fontSize',15);
set(gca,'LooseInset',get(gca,'TightInset'));
boxplot(distd,{sigmaExtText,LinProposal},'factorgap',[5 2],'labelverbosity','minor','datalim',[-Inf Inf]);
if minimalSetFlag==1 && lengthLinesFlag==0
    title('20m line length - Minimal Set');
end
if minimalSetFlag==1 && lengthLinesFlag==1
    title('1m line length - Minimal Set');
end

if minimalSetFlag==0 && lengthLinesFlag==0
    title('20m line length - All points');
end

if minimalSetFlag==0 && lengthLinesFlag==1
    title('1m line length - All points');
end
xlabel('gaussian error (pixels)');
ylabel('distance error (m)');

